@@ -1,8 +1,6 @@
 import decimal
 import logging
-import math
 import re
-from abc import ABC, abstractmethod, ABCMeta
 from typing import List, Tuple, Optional
 
 from flask import request
@@ -10,6 +8,7 @@ from flask_restx import Resource, Namespace, fields
 
 from ezbeq.catalogue import CatalogueProvider, CatalogueEntry
 from ezbeq.device import DeviceRepository, InvalidRequestError
+from ezbeq.iir import PeakingEQ, LowShelf, HighShelf
 
 logger = logging.getLogger('ezbeq.devices')
 
@@ -235,10 +234,10 @@ class Devices(Resource):
         self.__bridge: DeviceRepository = kwargs['device_bridge']
 
     def get(self):
-        return {n: d.serialise() for n, d in self.__bridge.all_devices().items()}
+        return {n: d.serialise() for n, d in self.__bridge.all_devices(refresh=True).items()}
 
 
-slot_model = v1_api.model('Slot', {
+slot_model_v1 = v1_api.model('Slot', {
     'id': fields.String(required=True),
     'active': fields.Boolean(required=False),
     'gain1': fields.Float(required=False),
@@ -248,10 +247,11 @@ slot_model = v1_api.model('Slot', {
     'entry': fields.String(required=False)
 })
 
-device_model = v1_api.model('Device', {
+
+device_model_v1 = v1_api.model('Device', {
     'mute': fields.Boolean(required=False),
     'masterVolume': fields.Float(required=False),
-    'slots': fields.List(fields.Nested(slot_model), required=False)
+    'slots': fields.List(fields.Nested(slot_model_v1), required=False)
 })
 
 
@@ -263,7 +263,40 @@ class Device(Resource):
         self.__bridge: DeviceRepository = kwargs['device_bridge']
         self.__catalogue_provider: CatalogueProvider = kwargs['catalogue']
 
-    @v1_api.expect(device_model, validate=True)
+    @v1_api.expect(device_model_v1, validate=True)
+    def patch(self, device_name: str):
+        payload = request.get_json()
+        logger.info(f"PATCHing {device_name} with {payload}")
+        if not self.__bridge.update(device_name, payload):
+            logger.info(f"PATCH {device_name} was a nop")
+        return self.__bridge.state(device_name).serialise()
+
+
+slot_model_v2 = v2_api.model('Slot', {
+    'id': fields.String(required=True),
+    'active': fields.Boolean(required=False),
+    'gains': fields.List(fields.Float, required=False),
+    'mutes': fields.List(fields.Boolean,required=False),
+    'entry': fields.String(required=False)
+})
+
+
+device_model_v2 = v2_api.model('Device', {
+    'mute': fields.Boolean(required=False),
+    'masterVolume': fields.Float(required=False),
+    'slots': fields.List(fields.Nested(slot_model_v1), required=False)
+})
+
+
+@v2_api.route('/<string:device_name>')
+class Device(Resource):
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.__bridge: DeviceRepository = kwargs['device_bridge']
+        self.__catalogue_provider: CatalogueProvider = kwargs['catalogue']
+
+    @v2_api.expect(device_model_v2, validate=True)
     def patch(self, device_name: str):
         payload = request.get_json()
         logger.info(f"PATCHing {device_name} with {payload}")
@@ -475,129 +508,3 @@ class DeviceSender(Resource):
 
     def delete(self, slot):
         return delete_filter(self.__bridge, 'master', slot)
-
-
-def float_to_str(f):
-    """
-    Convert the given float to a string without scientific notation or in the correct hex format.
-    """
-    d1 = ctx.create_decimal(repr(f))
-    return format(d1, 'f')
-
-
-class Biquad(ABC):
-
-    def __init__(self, fs, freq, q, gain):
-        self.fs = fs
-        self.gain = round(float(gain), 3)
-        self.freq = round(float(freq), 2)
-        self.q = round(float(q), 4)
-        self.w0 = 2.0 * math.pi * freq / fs
-        self.cos_w0 = math.cos(self.w0)
-        self.sin_w0 = math.sin(self.w0)
-        self.alpha = self.sin_w0 / (2.0 * self.q)
-        self.A = 10.0 ** (self.gain / 40.0)
-        self.a, self.b = self._compute_coeffs()
-
-    def __len__(self):
-        return 1
-
-    @abstractmethod
-    def _compute_coeffs(self):
-        pass
-
-    def format_biquads(self) -> dict:
-        a = {f"a{idx}": f"{float_to_str(-x)}" for idx, x in enumerate(self.a) if idx != 0}
-        b = {f"b{idx}": f"{float_to_str(x)}" for idx, x in enumerate(self.b)}
-        return {**b, **a}
-
-    def __repr__(self):
-        return f"{self.print_params()}|{'`'.join(self.format_biquads())}"
-
-    def print_params(self):
-        return f"{self.__class__.__name__}|{self.freq}|{self.gain}|{self.q}"
-
-
-class PeakingEQ(Biquad):
-    '''
-    H(s) = (s^2 + s*(A/Q) + 1) / (s^2 + s/(A*Q) + 1)
-
-            b0 =   1 + alpha*A
-            b1 =  -2*cos(w0)
-            b2 =   1 - alpha*A
-            a0 =   1 + alpha/A
-            a1 =  -2*cos(w0)
-            a2 =   1 - alpha/A
-    '''
-
-    def __init__(self, fs, freq, q, gain):
-        super().__init__(fs, freq, q, gain)
-
-    def _compute_coeffs(self):
-        A = self.A
-        a = [1.0 + self.alpha / A, -2.0 * self.cos_w0, 1.0 - self.alpha / A]
-        b = [1.0 + self.alpha * A, -2.0 * self.cos_w0, 1.0 - self.alpha * A]
-        return [a1 / a[0] for a1 in a], [b1 / a[0] for b1 in b]
-
-
-class LowShelf(Biquad):
-    '''
-    lowShelf: H(s) = A * (s^2 + (sqrt(A)/Q)*s + A)/(A*s^2 + (sqrt(A)/Q)*s + 1)
-
-            b0 =    A*( (A+1) - (A-1)*cos(w0) + 2*sqrt(A)*alpha )
-            b1 =  2*A*( (A-1) - (A+1)*cos(w0)                   )
-            b2 =    A*( (A+1) - (A-1)*cos(w0) - 2*sqrt(A)*alpha )
-            a0 =        (A+1) + (A-1)*cos(w0) + 2*sqrt(A)*alpha
-            a1 =   -2*( (A-1) + (A+1)*cos(w0)                   )
-            a2 =        (A+1) + (A-1)*cos(w0) - 2*sqrt(A)*alpha
-    '''
-
-    def __init__(self, fs, freq, q, gain):
-        super().__init__(fs, freq, q, gain)
-
-    def _compute_coeffs(self):
-        A = self.A
-        a = [
-            (A + 1) + ((A - 1) * self.cos_w0) + (2.0 * math.sqrt(A) * self.alpha),
-            -2.0 * ((A - 1) + ((A + 1) * self.cos_w0)),
-            (A + 1) + ((A - 1) * self.cos_w0) - (2.0 * math.sqrt(A) * self.alpha)
-        ]
-        b = [
-            A * ((A + 1) - ((A - 1) * self.cos_w0) + (2.0 * math.sqrt(A) * self.alpha)),
-            2.0 * A * ((A - 1) - ((A + 1) * self.cos_w0)),
-            A * ((A + 1) - ((A - 1) * self.cos_w0) - (2 * math.sqrt(A) * self.alpha))
-        ]
-        return [a1 / a[0] for a1 in a], [b1 / a[0] for b1 in b]
-
-
-class HighShelf(Biquad):
-    '''
-    highShelf: H(s) = A * (A*s^2 + (sqrt(A)/Q)*s + 1)/(s^2 + (sqrt(A)/Q)*s + A)
-
-                b0 =    A*( (A+1) + (A-1)*cos(w0) + 2*sqrt(A)*alpha )
-                b1 = -2*A*( (A-1) + (A+1)*cos(w0)                   )
-                b2 =    A*( (A+1) + (A-1)*cos(w0) - 2*sqrt(A)*alpha )
-                a0 =        (A+1) - (A-1)*cos(w0) + 2*sqrt(A)*alpha
-                a1 =    2*( (A-1) - (A+1)*cos(w0)                   )
-                a2 =        (A+1) - (A-1)*cos(w0) - 2*sqrt(A)*alpha
-
-    '''
-
-    def __init__(self, fs, freq, q, gain):
-        super().__init__(fs, freq, q, gain)
-
-    def _compute_coeffs(self):
-        A = self.A
-        cos_w0 = self.cos_w0
-        alpha = self.alpha
-        a = [
-            (A + 1) - ((A - 1) * cos_w0) + (2.0 * math.sqrt(A) * alpha),
-            2.0 * ((A - 1) - ((A + 1) * cos_w0)),
-            (A + 1) - ((A - 1) * cos_w0) - (2.0 * math.sqrt(A) * alpha)
-        ]
-        b = [
-            A * ((A + 1) + ((A - 1) * cos_w0) + (2.0 * math.sqrt(A) * alpha)),
-            -2.0 * A * ((A - 1) + ((A + 1) * cos_w0)),
-            A * ((A + 1) + ((A - 1) * cos_w0) - (2.0 * math.sqrt(A) * alpha))
-        ]
-        return [a1 / a[0] for a1 in a], [b1 / a[0] for b1 in b]
