@@ -1,12 +1,13 @@
 import json
 import logging
+import math
 import os
+import time
 from concurrent.futures.thread import ThreadPoolExecutor
 from contextlib import contextmanager
 from typing import List, Optional, Union
 
-import math
-import time
+import yaml
 from autobahn.exception import Disconnected
 from autobahn.twisted import WebSocketClientFactory, WebSocketClientProtocol
 from twisted.internet.protocol import ReconnectingClientFactory
@@ -15,18 +16,28 @@ from ezbeq.apis.ws import WsServer
 from ezbeq.catalogue import CatalogueEntry, CatalogueProvider
 from ezbeq.device import InvalidRequestError, SlotState, PersistentDevice, DeviceState
 
+INPUT_NAME = 'input'
+OUTPUT_NAME = 'output'
+CROSSOVER_NAME = 'crossover'
+
 logger = logging.getLogger('ezbeq.minidsp')
 
 
 class MinidspState(DeviceState):
 
-    def __init__(self, name: str, **kwargs):
+    def __init__(self, name: str, descriptor: 'MinidspDescriptor', **kwargs):
         self.__name = name
         self.master_volume: float = kwargs['mv'] if 'mv' in kwargs else 0.0
         self.__mute: bool = kwargs['mute'] if 'mute' in kwargs else False
         self.__active_slot: str = kwargs['active_slot'] if 'active_slot' in kwargs else ''
+        self.__descriptr = descriptor
         slot_ids = [str(i + 1) for i in range(4)]
-        self.__slots: List[MinidspSlotState] = [MinidspSlotState(c_id, c_id == self.active_slot) for c_id in slot_ids]
+        self.__slots: List[MinidspSlotState] = [
+            MinidspSlotState(c_id,
+                             c_id == self.active_slot,
+                             0 if not descriptor.input else len(descriptor.input.channels),
+                             0 if not descriptor.output else len(descriptor.output.channels)) for c_id in slot_ids
+        ]
 
     def update_master_state(self, mute: bool, gain: float):
         self.__mute = mute
@@ -99,30 +110,28 @@ class MinidspState(DeviceState):
 
 class MinidspSlotState(SlotState['MinidspSlotState']):
 
-    def __init__(self, slot_id: str, active: bool):
+    def __init__(self, slot_id: str, active: bool, input_channels: int, output_channels: int):
         super().__init__(slot_id)
-        self.gain1 = 0.0
-        self.mute1 = False
-        self.gain2 = 0.0
-        self.mute2 = False
+        self.__input_channels = input_channels
+        self.__output_channels = output_channels
+        self.gains = self.__make_vals(0.0)
+        self.mutes = self.__make_vals(False)
         self.active = active
 
     def clear(self):
         super().clear()
-        self.gain1 = 0.0
-        self.gain2 = 0.0
-        self.mute1 = False
-        self.mute2 = False
+        self.gains = self.__make_vals(0.0)
+        self.mutes = self.__make_vals(False)
+
+    def __make_vals(self, val):
+        return [val] * self.__input_channels
 
     def set_gain(self, channel: Optional[int], value: float):
         if channel is None:
-            self.gain1 = value
-            self.gain2 = value
+            self.gains = self.__make_vals(value)
         else:
-            if channel == 1:
-                self.gain1 = value
-            elif channel == 2:
-                self.gain2 = value
+            if channel <= self.__input_channels:
+                self.gains[channel-1] = value
             else:
                 raise ValueError(f'Unknown channel {channel} for slot {self.slot_id}')
 
@@ -131,13 +140,10 @@ class MinidspSlotState(SlotState['MinidspSlotState']):
 
     def __do_mute(self, channel: Optional[int], value: bool):
         if channel is None:
-            self.mute1 = value
-            self.mute2 = value
+            self.mutes = self.__make_vals(value)
         else:
-            if channel == 1:
-                self.mute1 = value
-            elif channel == 2:
-                self.mute2 = value
+            if channel <= self.__input_channels:
+                self.mutes[channel-1] = value
             else:
                 raise ValueError(f'Unknown channel {channel} for slot {self.slot_id}')
 
@@ -146,28 +152,235 @@ class MinidspSlotState(SlotState['MinidspSlotState']):
 
     def merge_with(self, state: dict) -> None:
         super().merge_with(state)
-        if 'gain1' in state:
-            self.gain1 = float(state['gain1'])
-        if 'gain2' in state:
-            self.gain2 = float(state['gain2'])
-        if 'mute1' in state:
-            self.mute1 = bool(state['mute1'])
-        if 'mute2' in state:
-            self.mute2 = bool(state['mute2'])
+        # legacy (v1 api)
+        if 'gain1' in state and self.__input_channels > 0:
+            self.gains[0] = float(state['gain1'])
+        if 'gain2' in state and self.__input_channels > 1:
+            self.gains[1] = float(state['gain2'])
+        if 'mute1' in state and self.__input_channels > 0:
+            self.mutes[0] = bool(state['mute1'])
+        if 'mute2' in state and self.__input_channels > 1:
+            self.mutes[1] = bool(state['mute2'])
+        # current (v2 api)
+        if 'gains' in state and len(state['gains']) == self.__input_channels:
+            self.gains = [float(v) for v in state['gains']]
+        if 'mutes' in state and len(state['mutes']) == self.__input_channels:
+            self.mutes = [bool(v) for v in state['mutes']]
 
     def as_dict(self) -> dict:
         sup = super().as_dict()
+        vals = {}
+        if self.__input_channels == 2:
+            # backwards compatibility
+            vals = {
+                'gain1': self.gains[0],
+                'gain2': self.gains[1],
+                'mute1': self.mutes[0],
+                'mute2': self.mutes[1],
+            }
         return {
             **sup,
-            'gain1': self.gain1,
-            'gain2': self.gain2,
-            'mute1': self.mute1,
-            'mute2': self.mute2,
-            'canActivate': True
+            **vals,
+            'gains': [g for g in self.gains],
+            'mutes': [m for m in self.mutes],
+            'canActivate': True,
+            'inputs': self.__input_channels,
+            'outputs': self.__output_channels
         }
 
     def __repr__(self):
-        return f"{super().__repr__()} - 1: {self.gain1:.2f}/{self.mute1} 2: {self.gain2:.2f}/{self.mute2}"
+        vals = ' '.join([f"{i+1}: {g:.2f}/{self.mutes[i]}" for i, g in enumerate(self.gains)])
+        return f"{super().__repr__()} - {vals}"
+
+
+class PeqRoutes:
+    def __init__(self, name: str, biquads: int, channels: List[int], beq_slots: List[int], groups: List[int] = None):
+        self.name = name
+        self.biquads = biquads
+        self.channels = channels
+        self.beq_slots = beq_slots
+        self.groups = groups
+
+    @property
+    def takes_beq(self) -> bool:
+        return len(self.channels) > 0 and len(self.beq_slots) > 0
+
+    def __repr__(self):
+        return f"{self.name}"
+
+
+class BeqFilterSlot:
+
+    def __init__(self, name: str, idx: int, channels: List[int], group: Optional[int] = None):
+        self.name = name
+        self.idx = idx
+        self.channels = channels
+        self.group = group
+
+    def __repr__(self):
+        return f"{self.name}{self.group if self.group is not None else ''}/{self.idx}/{self.channels}"
+
+
+class BeqFilterAllocator:
+
+    def __init__(self, routes: List[PeqRoutes]):
+        self.slots = []
+        for r in routes:
+            if r:
+                for s in r.beq_slots:
+                    if r.groups:
+                        for g in r.groups:
+                            self.slots.append(BeqFilterSlot(r.name, s, r.channels, g))
+                    else:
+                        self.slots.append(BeqFilterSlot(r.name, s, r.channels))
+
+    def pop(self) -> Optional[BeqFilterSlot]:
+        if self.slots:
+            return self.slots.pop(0)
+        return None
+
+    def __len__(self):
+        return len(self.slots)
+
+    def __repr__(self):
+        return f"{self.slots}"
+
+
+class MinidspDescriptor:
+
+    def __init__(self, name: str, fs: str, i: PeqRoutes = None, xo: PeqRoutes = None, o: PeqRoutes = None,
+                 extra: List[PeqRoutes] = None):
+        self.name = name
+        self.fs = str(int(fs))
+        self.input = i
+        self.crossover = xo
+        self.output = o
+        self.extra = extra
+
+    @property
+    def peq_routes(self) -> List[PeqRoutes]:
+        return [x for x in [self.input, self.crossover, self.output, self.extra] if x]
+
+    def to_allocator(self) -> BeqFilterAllocator:
+        return BeqFilterAllocator(self.peq_routes)
+
+    def __repr__(self):
+        s = f"{self.name}, fs:{self.fs}"
+        if self.input:
+            s = f"{s}, inputs: {self.input}"
+        if self.crossover:
+            s = f"{s}, crossovers: {self.crossover}"
+        if self.output:
+            s = f"{s}, outputs: {self.output}"
+        return s
+
+
+def zero_til(count: int) -> List[int]:
+    return list(range(0, count))
+
+
+class Minidsp24HD(MinidspDescriptor):
+
+    def __init__(self):
+        super().__init__('2x4HD',
+                         '96000',
+                         i=PeqRoutes(INPUT_NAME, 10, zero_til(2), zero_til(10)),
+                         xo=PeqRoutes(CROSSOVER_NAME, 4, zero_til(4), [], groups=zero_til(2)),
+                         o=PeqRoutes(OUTPUT_NAME, 10, zero_til(4), []))
+
+
+class MinidspDDRC24(MinidspDescriptor):
+
+    def __init__(self):
+        super().__init__('DDRC24',
+                         '48000',
+                         xo=PeqRoutes(CROSSOVER_NAME, 4, zero_til(4), [], zero_til(2)),
+                         o=PeqRoutes(OUTPUT_NAME, 10, zero_til(4), zero_til(10)))
+
+
+class MinidspDDRC88(MinidspDescriptor):
+
+    def __init__(self, sw_channels: List[int] = None):
+        c = sw_channels if sw_channels is not None else [3]
+        if any(ch for ch in c if ch < 0 or ch > 7):
+            raise ValueError(f"Invalid channels {c}")
+        non_sw = [c1 for c1 in zero_til(8) if c1 not in c]
+        super().__init__('DDRC88',
+                         '48000',
+                         xo=PeqRoutes(CROSSOVER_NAME, 8, zero_til(8), [], zero_til(2)),
+                         o=PeqRoutes(OUTPUT_NAME, 10, c, zero_til(10)),
+                         extra=PeqRoutes(OUTPUT_NAME, 10, non_sw, []) if non_sw else None)
+
+
+class Minidsp410(MinidspDescriptor):
+
+    def __init__(self):
+        super().__init__('4x10',
+                         '96000',
+                         i=PeqRoutes(INPUT_NAME, 5, zero_til(2), zero_til(5)),
+                         o=PeqRoutes(OUTPUT_NAME, 5, zero_til(8), zero_til(5)))
+
+
+class Minidsp1010(MinidspDescriptor):
+
+    def __init__(self, use_xo: Union[bool, int, str]):
+        if use_xo is True:
+            secondary = {'xo': PeqRoutes(CROSSOVER_NAME, 4, zero_til(8), zero_til(4), groups=[0])}
+        elif use_xo is False:
+            secondary = {'o': PeqRoutes(OUTPUT_NAME, 6, zero_til(8), zero_til(4))}
+        elif use_xo == '0' or use_xo == '1':
+            secondary = {'xo': PeqRoutes(CROSSOVER_NAME, 4, zero_til(8), zero_til(4), groups=[int(use_xo)])}
+        elif use_xo == 0 or use_xo == 1:
+            secondary = {'xo': PeqRoutes(CROSSOVER_NAME, 4, zero_til(8), zero_til(4), groups=[use_xo])}
+        elif use_xo == 'all':
+            secondary = {'xo': PeqRoutes(CROSSOVER_NAME, 4, zero_til(8), zero_til(4), groups=zero_til(2))}
+        else:
+            secondary = {'o': PeqRoutes(OUTPUT_NAME, 6, zero_til(8), zero_til(4))}
+        super().__init__('10x10',
+                         '48000',
+                         i=PeqRoutes(INPUT_NAME, 6, zero_til(8), zero_til(6)),
+                         **secondary)
+
+
+def make_peq_layout(cfg: dict) -> MinidspDescriptor:
+    if 'device_type' in cfg:
+        device_type = cfg['device_type']
+        if device_type == '24HD':
+            return Minidsp24HD()
+        elif device_type == 'DDRC24':
+            return MinidspDDRC24()
+        elif device_type == 'DDRC88':
+            return MinidspDDRC88(sw_channels=cfg.get('sw_channels', None))
+        elif device_type == '4x10':
+            return Minidsp410()
+        elif device_type == '10x10':
+            return Minidsp1010(cfg.get('use_xo', False))
+        elif device_type == 'SHD':
+            return MinidspDDRC24()
+    elif 'descriptor' in cfg:
+        desc: dict = cfg['descriptor']
+        named_args = ['name', 'fs', 'routes']
+        missing_keys = [x for x in named_args if x not in desc.keys()]
+        if missing_keys:
+            raise ValueError(f"Custom descriptor is missing keys - {missing_keys} - from {desc}")
+        routes: List[dict] = desc['routes']
+
+        def make_route(r: dict) -> PeqRoutes:
+            r_named_args = ['name', 'biquads', 'channels', 'slots']
+            missing_route_keys = [x for x in r_named_args if x not in r.keys()]
+            if missing_route_keys:
+                raise ValueError(f"Custom PeqRoutes is missing keys - {missing_keys} - from {r}")
+
+            def to_ints(v):
+                return [int(i) for i in v] if v else None
+
+            return PeqRoutes(r['name'], int(r['biquads']), to_ints(r['channels']), to_ints(r['slots']),
+                             to_ints(r.get('groups', None)))
+        args = [make_route(r) for r in routes]
+        return MinidspDescriptor(desc['name'], str(desc['fs']),
+                                 **{'xo' if r.name == CROSSOVER_NAME else r.name[0]: r for r in args})
+    else:
+        return Minidsp24HD()
 
 
 class Minidsp(PersistentDevice[MinidspState]):
@@ -178,9 +391,13 @@ class Minidsp(PersistentDevice[MinidspState]):
         self.__executor = ThreadPoolExecutor(max_workers=1)
         self.__cmd_timeout = cfg.get('cmdTimeout', 10)
         self.__ignore_retcode = cfg.get('ignoreRetcode', False)
+        self.__slot_change_delay: Union[bool, int] = cfg.get('slotChangeDelay', False)
         self.__levels_interval = 1.0 / float(cfg.get('levelsFps', 10))
         self.__runner = cfg['make_runner']()
         self.__client = MinidspRsClient(self) if cfg.get('useWs', False) else None
+        self.__descriptor: MinidspDescriptor = make_peq_layout(cfg)
+        logger.info("Minidsp descriptor is loaded....")
+        logger.info(yaml.dump(self.__descriptor, indent=2, default_flow_style=False, sort_keys=False))
         ws_server.factory.set_levels_provider(name, self.start_broadcast_levels)
 
     @property
@@ -193,7 +410,7 @@ class Minidsp(PersistentDevice[MinidspState]):
 
     def __load_state(self) -> MinidspState:
         result = self.__executor.submit(self.__read_state_from_device).result(timeout=self.__cmd_timeout)
-        return result if result else MinidspState(self.name)
+        return result if result else MinidspState(self.name, self.__descriptor)
 
     def __read_state_from_device(self) -> Optional[MinidspState]:
         output = None
@@ -208,7 +425,7 @@ class Minidsp(PersistentDevice[MinidspState]):
                     'mute': status['master']['mute'],
                     'mv': status['master']['volume']
                 }
-                return MinidspState(self.name, **values)
+                return MinidspState(self.name, self.__descriptor, **values)
             else:
                 logger.error(f"No output returned from device")
         except:
@@ -220,7 +437,7 @@ class Minidsp(PersistentDevice[MinidspState]):
         return int(idx) - 1
 
     def __send_cmds(self, target_slot_idx: Optional[int], cmds: List[str]):
-        return self.__executor.submit(self.__do_run, cmds, target_slot_idx).result(timeout=self.__cmd_timeout)
+        return self.__executor.submit(self.__do_run, cmds, target_slot_idx, self.__slot_change_delay).result(timeout=self.__cmd_timeout)
 
     def activate(self, slot: str):
         def __do_it():
@@ -236,7 +453,8 @@ class Minidsp(PersistentDevice[MinidspState]):
         if target_slot_idx < 0 or target_slot_idx > 3:
             raise InvalidRequestError(f"Slot must be in range 1-4")
 
-    def load_biquads(self, slot: str, overwrite: bool, inputs: List[int], outputs: List[int], biquads: List[dict]) -> None:
+    def load_biquads(self, slot: str, overwrite: bool, inputs: List[int], outputs: List[int],
+                     biquads: List[dict]) -> None:
         def __do_it():
             target_slot_idx = self.__as_idx(slot)
             self.__validate_slot_idx(target_slot_idx)
@@ -257,7 +475,7 @@ class Minidsp(PersistentDevice[MinidspState]):
         def __do_it():
             target_slot_idx = self.__as_idx(slot)
             self.__validate_slot_idx(target_slot_idx)
-            cmds = MinidspBeqCommandGenerator.filt(entry)
+            cmds = MinidspBeqCommandGenerator.filt(entry, self.__descriptor)
             try:
                 self.__send_cmds(target_slot_idx, cmds)
                 self._current_state.load(slot, entry.formatted_title)
@@ -271,9 +489,20 @@ class Minidsp(PersistentDevice[MinidspState]):
         def __do_it():
             target_slot_idx = self.__as_idx(slot)
             self.__validate_slot_idx(target_slot_idx)
-            cmds = MinidspBeqCommandGenerator.filt(None)
-            cmds.extend(MinidspBeqCommandGenerator.mute(False, target_slot_idx, None))
-            cmds.extend(MinidspBeqCommandGenerator.gain(0.0, target_slot_idx, None))
+            cmds = MinidspBeqCommandGenerator.filt(None, self.__descriptor)
+            beq_slots = self.__descriptor.to_allocator()
+            levels = []
+            handled = []
+            s = beq_slots.pop()
+            while s is not None:
+                for c in s.channels:
+                    if s.name == INPUT_NAME and c not in handled:
+                        levels.extend(MinidspBeqCommandGenerator.mute(False, target_slot_idx, c, side=s.name))
+                        levels.extend(MinidspBeqCommandGenerator.gain(0.0, target_slot_idx, c, side=s.name))
+                        handled.append(c)
+                s = beq_slots.pop()
+            if levels:
+                cmds.extend(levels)
             try:
                 self.__send_cmds(target_slot_idx, cmds)
                 self._current_state.clear(slot)
@@ -314,15 +543,23 @@ class Minidsp(PersistentDevice[MinidspState]):
         target_channel_idx = self.__as_idx(channel) if channel else None
         return target_channel_idx, target_slot_idx
 
-    def __do_run(self, config_cmds: List[str], slot: Optional[int]):
+    def __do_run(self, config_cmds: List[str], slot: Optional[int], slot_change_delay: Union[bool, int]):
         if slot is not None:
             change_slot = True
             current_state = self.__read_state_from_device()
             if current_state and current_state.active_slot == str(slot + 1):
                 change_slot = False
             if change_slot is True:
-                logger.info(f"Activating slot {slot}, current is {current_state.active_slot if current_state else 'UNKNOWN'}")
-                config_cmds.insert(0, MinidspBeqCommandGenerator.activate(slot))
+                if slot_change_delay:
+                    self.__do_run([], slot, False)
+                    if slot_change_delay is not True and slot_change_delay > 0:
+                        from time import sleep
+                        logger.info(f"Sleeping for {slot_change_delay} seconds after config slot change")
+                        sleep(slot_change_delay)
+                else:
+                    logger.info(
+                        f"Activating slot {slot}, current is {current_state.active_slot if current_state else 'UNKNOWN'}")
+                    config_cmds.insert(0, MinidspBeqCommandGenerator.activate(slot))
         formatted = '\n'.join(config_cmds)
         logger.info(f"\n{formatted}")
         with tmp_file(config_cmds) as file_name:
@@ -338,8 +575,8 @@ class Minidsp(PersistentDevice[MinidspState]):
     def _load_initial_state(self) -> MinidspState:
         return self.__load_state()
 
-    def state(self) -> MinidspState:
-        if not self._hydrate():
+    def state(self, refresh: bool = False) -> MinidspState:
+        if not self._hydrate() or refresh is True:
             new_state = self.__load_state()
             self._current_state.update_master_state(new_state.mute, new_state.master_volume)
         return self._current_state
@@ -370,6 +607,7 @@ class Minidsp(PersistentDevice[MinidspState]):
     def __update_slot(self, slot: dict) -> bool:
         any_update = False
         current_slot = self._current_state.get_slot(slot['id'])
+        # legacy
         if 'gain1' in slot:
             self.set_gain(current_slot.slot_id, 1, slot['gain1'])
             any_update = True
@@ -388,6 +626,18 @@ class Minidsp(PersistentDevice[MinidspState]):
             else:
                 self.unmute(current_slot.slot_id, 2)
             any_update = True
+        # current
+        if 'gains' in slot:
+            for idx, gain in enumerate(slot['gains']):
+                self.set_gain(current_slot.slot_id, idx+1, gain)
+                any_update = True
+        if 'mutes' in slot:
+            for idx, mute in enumerate(slot['mutes']):
+                if mute is True:
+                    self.mute(current_slot.slot_id, idx+1)
+                else:
+                    self.unmute(current_slot.slot_id, idx+1)
+                any_update = True
         if 'entry' in slot:
             if slot['entry']:
                 match = self.__catalogue.find(slot['entry'])
@@ -416,8 +666,8 @@ class Minidsp(PersistentDevice[MinidspState]):
             logger.info(f"readlevels,{ts},{to_millis(start, end)}")
             return {
                 'ts': ts,
-                'input': levels['input_levels'],
-                'output': levels['output_levels']
+                INPUT_NAME: levels['input_levels'],
+                OUTPUT_NAME: levels['output_levels']
             }
         except:
             logger.exception(f"Unable to load levels {lines}")
@@ -446,7 +696,7 @@ class MinidspBeqCommandGenerator:
         # [in|out]put <channel> peq <index> set -- <b0> <b1> <b2> <a1> <a2>
         # [in|out]put <channel> peq <index> bypass [on|off]
         cmds = []
-        for side, channels in {'input': inputs, 'output': outputs}.items():
+        for side, channels in {INPUT_NAME: inputs, OUTPUT_NAME: outputs}.items():
             for channel in channels:
                 for idx, bq in enumerate(biquads):
                     if bq:
@@ -462,73 +712,112 @@ class MinidspBeqCommandGenerator:
         return cmds
 
     @staticmethod
-    def filt(entry: Optional[CatalogueEntry]):
-        # input <channel> peq <index> set -- <b0> <b1> <b2> <a1> <a2>
-        # input <channel> peq <index> bypass [on|off]
+    def as_bq(f: dict, fs: str):
+        if fs in f['biquads']:
+            bq = f['biquads'][fs]['b'] + f['biquads'][fs]['a']
+        else:
+            t = f['type']
+            freq = f['freq']
+            gain = f['gain']
+            q = f['q']
+            from ezbeq.iir import PeakingEQ, LowShelf, HighShelf
+            if t == 'PeakingEQ':
+                f = PeakingEQ(int(fs), freq, q, gain)
+            elif t == 'LowShelf':
+                f = LowShelf(int(fs), freq, q, gain)
+            elif t == 'HighShelf':
+                f = HighShelf(int(fs), freq, q, gain)
+            else:
+                raise InvalidRequestError(f"Unknown filt_type {t}")
+            bq = list(f.format_biquads().values())
+        if len(bq) != 5:
+            raise ValueError(f"Invalid coeff count {len(bq)}")
+        return bq
+
+    @staticmethod
+    def filt(entry: Optional[CatalogueEntry], descriptor: MinidspDescriptor):
+        # [in|out]put <channel> peq <index> set -- <b0> <b1> <b2> <a1> <a2>
+        # [in|out]put <channel> peq <index> bypass [on|off]
         cmds = []
-        for c in range(2):
-            idx = 0
-            if entry:
-                for f in entry.filters:
-                    bq: dict = f['biquads']['96000']
-                    coeffs: List[str] = bq['b'] + bq['a']
-                    if len(coeffs) != 5:
-                        raise ValueError(f"Invalid coeff count {len(coeffs)} at idx {idx}")
-                    else:
-                        cmds.append(MinidspBeqCommandGenerator.bq(c, idx, coeffs))
-                        cmds.append(MinidspBeqCommandGenerator.bypass(c, idx, False))
-                        idx += 1
-            for i in range(idx, 10):
-                cmds.append(MinidspBeqCommandGenerator.bypass(c, i, True))
+        # write filts to the inputs first then the output if it's a split device
+        filters = [MinidspBeqCommandGenerator.as_bq(f, descriptor.fs) for f in entry.filters] if entry else []
+        beq_slots = descriptor.to_allocator()
+
+        def push(chs: List[int], i: int, s: str, group: Optional[int]):
+            for ch in chs:
+                cmds.append(MinidspBeqCommandGenerator.bq(ch, i, coeffs, s, group=group))
+                cmds.append(MinidspBeqCommandGenerator.bypass(ch, i, False, s, group=group))
+
+        idx = 0
+        while idx < len(filters):
+            coeffs: List[str] = filters[idx]
+            slot = beq_slots.pop()
+            if slot is not None:
+                push(slot.channels, slot.idx, slot.name, slot.group)
+            else:
+                raise ValueError(f"Loaded {idx} filters but no slots remaining")
+            idx += 1
+        s = beq_slots.pop()
+        while s is not None:
+            for c in s.channels:
+                cmds.append(MinidspBeqCommandGenerator.bypass(c, s.idx, True, s.name, s.group))
+            s = beq_slots.pop()
         return cmds
 
     @staticmethod
-    def bq(channel: int, idx: int, coeffs, side: str = 'input'):
-        return f"{side} {channel} peq {idx} set -- {' '.join(coeffs)}"
+    def bq(channel: int, idx: int, coeffs, side: str = INPUT_NAME, group: Optional[int] = None):
+        is_xo = side == CROSSOVER_NAME
+        addr = f"crossover {group}" if is_xo and group is not None else 'peq'
+        return f"{OUTPUT_NAME if is_xo else side} {channel} {addr} {idx} set -- {' '.join(coeffs)}"
 
     @staticmethod
-    def bypass(channel: int, idx: int, bypass: bool, side: str = 'input'):
-        return f"{side} {channel} peq {idx} bypass {'on' if bypass else 'off'}"
+    def bypass(channel: int, idx: int, bypass: bool, side: str = INPUT_NAME, group: Optional[int] = 0):
+        is_xo = side == CROSSOVER_NAME
+        addr = f"crossover {group}" if is_xo and group is not None else 'peq'
+        return f"{OUTPUT_NAME if is_xo else side} {channel} {addr} {idx} bypass {'on' if bypass else 'off'}"
 
     @staticmethod
-    def mute(state: bool, slot: Optional[int], channel: Optional[int]):
+    def mute(state: bool, slot: Optional[int], channel: Optional[int], side: Optional[str] = INPUT_NAME):
         '''
         Generates commands to mute the configuration.
         :param state: mute if true otherwise unmute.
         :param slot: the target slot, if not set apply to the master control.
         :param channel: the channel, applicable only if slot is set, if not set apply to both input channels.
+        :param side: the side, input by default.
         :return: the commands.
         '''
         state_cmd = 'on' if state else 'off'
         if slot is not None:
             cmds = []
             if channel is None:
-                cmds.append(f"input 0 mute {state_cmd}")
-                cmds.append(f"input 1 mute {state_cmd}")
+                cmds.append(f"{side} 0 mute {state_cmd}")
+                cmds.append(f"{side} 1 mute {state_cmd}")
             else:
-                cmds.append(f"input {channel} mute {state_cmd}")
+                cmds.append(f"{side} {channel} mute {state_cmd}")
             return cmds
         else:
             return [f"mute {state_cmd}"]
 
     @staticmethod
-    def gain(gain: float, slot: Optional[int], channel: Optional[int]):
+    def gain(gain: float, slot: Optional[int], channel: Optional[int], side: Optional[str] = INPUT_NAME):
         '''
         Generates commands to set gain.
         :param gain: the gain to set.
         :param slot: the target slot, if not set apply to the master control.
         :param channel: the channel, applicable only if slot is set, if not set apply to both input channels.
+        :param side: the side to apply the gain to, input by default.
         :return: the commands.
         '''
         if slot is not None:
+            # TODO is this valid for other devices
             if not -72.0 <= gain <= 12.0:
-                raise InvalidRequestError(f"Input gain {gain:.2f} out of range (>= -72.0 and <= 12.0)")
+                raise InvalidRequestError(f"{side} gain {gain:.2f} out of range (>= -72.0 and <= 12.0)")
             cmds = []
             if channel is None:
-                cmds.append(f"input 0 gain -- {gain:.2f}")
-                cmds.append(f"input 1 gain -- {gain:.2f}")
+                cmds.append(f"{side} 0 gain -- {gain:.2f}")
+                cmds.append(f"{side} 1 gain -- {gain:.2f}")
             else:
-                cmds.append(f"input {channel} gain -- {gain:.2f}")
+                cmds.append(f"{side} {channel} gain -- {gain:.2f}")
             return cmds
         else:
             if not -127.0 <= gain <= 0.0:
@@ -584,7 +873,7 @@ class MinidspRsProtocol(WebSocketClientProtocol):
 
     def onConnect(self, response):
         logger.info(f"Connected to {response.peer}")
-        self.sendMessage('getmso'.encode('utf-8'), isBinary=False)
+        # self.sendMessage('getmso'.encode('utf-8'), isBinary=False)
 
     def onOpen(self):
         logger.info("Connected to Minidsp")
